@@ -3,7 +3,7 @@ import json
 import os
 
 import pytest
-import requests
+import requests  # type: ignore
 
 from ...data_classes import ModelInformation
 from ...detector.detector_node import DetectorNode
@@ -93,10 +93,10 @@ async def test_sio_upload(test_detector_node: DetectorNode, sio_client):
 
 # NOTE: This test seems to be flaky.
 async def test_about_endpoint(test_detector_node: DetectorNode):
-    await asyncio.sleep(3)
+    await asyncio.sleep(16)
     response = requests.get(f'http://localhost:{GLOBALS.detector_port}/about', timeout=30)
 
-    assert response.status_code == 200
+    assert response.status_code == 200, response.content
     response_dict = json.loads(response.content)
     assert response_dict['model_info']
     model_information = ModelInformation.from_dict(response_dict['model_info'])
@@ -108,59 +108,60 @@ async def test_about_endpoint(test_detector_node: DetectorNode):
 
 
 async def test_model_version_api(test_detector_node: DetectorNode):
-    await asyncio.sleep(3)
+    async def await_correct_response(target_values: dict) -> None:
+        response_dict = {}
+        for _ in range(20):
+            await asyncio.sleep(1)
+            response = requests.get(f'http://localhost:{GLOBALS.detector_port}/model_version', timeout=30)
+            if not response.status_code == 200:
+                continue
+            response_dict = json.loads(response.content)
+            for key, target_value in target_values.items():
+                if key == 'local_versions':
+                    target_value = set(target_value)
 
-    response = requests.get(f'http://localhost:{GLOBALS.detector_port}/model_version', timeout=30)
-    assert response.status_code == 200
-    response_dict = json.loads(response.content)
-    assert response_dict['current_version'] == '1.1'
-    assert response_dict['target_version'] == '1.1'
-    assert response_dict['loop_version'] == '1.1'
-    assert response_dict['local_versions'] == ['1.1']
-    assert response_dict['version_control'] == 'follow_loop'
+                response_value = response_dict.get(key, None)
+                if response_value != target_value:
+                    break
+            return
+        raise Exception(f'Did not receive correct response: {response_dict} != {target_values}')
+
+    await await_correct_response({'version_control': 'follow_loop',
+                                  'current_version': '1.1',
+                                  'target_version': '1.1',
+                                  'loop_version': '1.1',
+                                  'local_versions': ['1.1']})
 
     response = requests.put(f'http://localhost:{GLOBALS.detector_port}/model_version', data='1.0', timeout=30)
-    response = requests.get(f'http://localhost:{GLOBALS.detector_port}/model_version', timeout=30)
-    assert response.status_code == 200
-    response_dict = json.loads(response.content)
-    assert response_dict['current_version'] == '1.1'
-    assert response_dict['target_version'] == '1.0'
-    assert response_dict['loop_version'] == '1.1'
-    assert response_dict['local_versions'] == ['1.1']
-    assert response_dict['version_control'] == 'specific_version'
+    assert response.status_code == 200, response.content
 
-    await asyncio.sleep(11)
+    await await_correct_response({'version_control': 'specific_version',
+                                  'current_version': '1.1',
+                                  'target_version': '1.0',
+                                  'loop_version': '1.1'})
 
-    response = requests.get(f'http://localhost:{GLOBALS.detector_port}/model_version', timeout=30)
-    assert response.status_code == 200
-    response_dict = json.loads(response.content)
-    assert response_dict['current_version'] == '1.0'
-    assert response_dict['target_version'] == '1.0'
-    assert response_dict['loop_version'] == '1.1'
-    assert set(response_dict['local_versions']) == set(['1.1', '1.0'])
-    assert response_dict['version_control'] == 'specific_version'
+    await await_correct_response({'version_control': 'specific_version',
+                                  'current_version': '1.0',
+                                  'target_version': '1.0',
+                                  'loop_version': '1.1',
+                                  'local_versions': ['1.1', '1.0']})
 
     response = requests.put(f'http://localhost:{GLOBALS.detector_port}/model_version', data='pause', timeout=30)
-    await asyncio.sleep(11)
-    response = requests.get(f'http://localhost:{GLOBALS.detector_port}/model_version', timeout=30)
-    assert response.status_code == 200
-    response_dict = json.loads(response.content)
-    assert response_dict['current_version'] == '1.0'
-    assert response_dict['target_version'] == '1.0'
-    assert response_dict['loop_version'] == '1.1'
-    assert set(response_dict['local_versions']) == set(['1.1', '1.0'])
-    assert response_dict['version_control'] == 'pause'
+    assert response.status_code == 200, response.content
+
+    await await_correct_response({'version_control': 'pause',
+                                  'current_version': '1.0',
+                                  'target_version': '1.0',
+                                  'loop_version': '1.1',
+                                  'local_versions': ['1.1', '1.0']})
 
     response = requests.put(f'http://localhost:{GLOBALS.detector_port}/model_version', data='follow_loop', timeout=30)
     await asyncio.sleep(11)
-    response = requests.get(f'http://localhost:{GLOBALS.detector_port}/model_version', timeout=30)
-    assert response.status_code == 200
-    response_dict = json.loads(response.content)
-    assert response_dict['current_version'] == '1.1'
-    assert response_dict['target_version'] == '1.1'
-    assert response_dict['loop_version'] == '1.1'
-    assert set(response_dict['local_versions']) == set(['1.1', '1.0'])
-    assert response_dict['version_control'] == 'follow_loop'
+    await await_correct_response({'version_control': 'follow_loop',
+                                  'current_version': '1.1',
+                                  'target_version': '1.1',
+                                  'loop_version': '1.1',
+                                  'local_versions': ['1.1', '1.0']})
 
 
 async def test_rest_outbox_mode(test_detector_node: DetectorNode):
@@ -169,9 +170,9 @@ async def test_rest_outbox_mode(test_detector_node: DetectorNode):
     def check_switch_to_mode(mode: str):
         response = requests.put(f'http://localhost:{GLOBALS.detector_port}/outbox_mode',
                                 data=mode, timeout=30)
-        assert response.status_code == 200
+        assert response.status_code == 200, response.content
         response = requests.get(f'http://localhost:{GLOBALS.detector_port}/outbox_mode', timeout=30)
-        assert response.status_code == 200
+        assert response.status_code == 200, response.content
         assert response.content == mode.encode()
 
     check_switch_to_mode('stopped')
@@ -185,7 +186,7 @@ async def test_api_responsive_during_large_upload(test_detector_node: DetectorNo
     with open(test_image_path, 'rb') as f:
         image_bytes = f.read()
 
-    for _ in range(100):
+    for _ in range(200):
         test_detector_node.outbox.save(image_bytes)
 
     outbox_size_early = len(get_outbox_files(test_detector_node.outbox))
@@ -193,8 +194,9 @@ async def test_api_responsive_during_large_upload(test_detector_node: DetectorNo
 
     # check if api is still responsive
     response = requests.get(f'http://localhost:{GLOBALS.detector_port}/outbox_mode', timeout=2)
-    assert response.status_code == 200
+    assert response.status_code == 200, response.content
 
     await asyncio.sleep(5)
     outbox_size_late = len(get_outbox_files(test_detector_node.outbox))
-    assert outbox_size_early > outbox_size_late > 0, 'The outbox should have been partially emptied.'
+    assert outbox_size_late > 0, 'The outbox should not be fully cleared, maybe the node was too fast.'
+    assert outbox_size_early > outbox_size_late, 'The outbox should have been partially emptied.'
