@@ -10,15 +10,18 @@ from typing import TYPE_CHECKING, Callable, Coroutine, Dict, List, Optional
 
 from fastapi.encoders import jsonable_encoder
 
-from ..data_classes import (Context, Errors, Hyperparameter, PretrainedModel, TrainerState, Training, TrainingData,
-                            TrainingOut, TrainingStateData)
-from ..helpers.misc import create_project_folder, delete_all_training_folders, generate_training, is_valid_uuid4
+from ..data_classes import (Context, Errors, PretrainedModel, TrainerState, Training, TrainingOut, TrainingStateData,
+                            TrainingStatus)
+from ..helpers.misc import create_project_folder, delete_all_training_folders, is_valid_uuid4
 from .downloader import TrainingsDownloader
+from .exceptions import CriticalError, NodeNeedsRestartError
 from .io_helpers import ActiveTrainingIO, EnvironmentVars, LastTrainingIO
 
 if TYPE_CHECKING:
     from .trainer_node import TrainerNode
 
+logger = logging.getLogger('learning_loop_node.trainer_logic_generic')
+
 
 class TrainerLogicGeneric(ABC):
 
@@ -63,19 +66,12 @@ class TrainerLogicGeneric(ABC):
         return self._training
 
     @property
-    def hyperparameter(self) -> Hyperparameter:
-        assert self.training_data is not None, 'Training should have data'
-        assert self.training_data.hyperparameter is not None, 'Training.data should have hyperparameter'
-        return self.training_data.hyperparameter
+    def hyperparameters(self) -> dict:
+        assert self._training is not None, 'Training should have data'
+        return self._training.hyperparameters
 
     # ---------------------------------------- PROPERTIES ----------------------------------------
 
-    @property
-    def training_data(self) -> Optional[TrainingData]:
-        if self.training_active and self.training.data:
-            return self.training.data
-        return None
-
     @property
     def training_context(self) -> Optional[Context]:
         if self.training_active:
@@ -108,12 +104,8 @@ class TrainerLogicGeneric(ABC):
     def hyperparameters_for_state_sync(self) -> Optional[Dict]:
         """Used in sync_confusion_matrix and send_status to provide information about the training configuration.
         """
-        if self._training and self._training.data and self._training.data.hyperparameter:
-            information = {}
-            information['resolution'] = self._training.data.hyperparameter.resolution
-            information['flipRl'] = self._training.data.hyperparameter.flip_rl
-            information['flipUd'] = self._training.data.hyperparameter.flip_ud
-            return information
+        if self._training:
+            return self._training.hyperparameters
         return None
 
     @property
@@ -170,13 +162,31 @@ class TrainerLogicGeneric(ABC):
         # Initializing a new training object will create the folder structure for the training.
         # The training loop will then run through the states of the training.
 
+    def generate_status_for_loop(self, trainer_uuid: str, trainer_name: str) -> TrainingStatus:
+
+        status = TrainingStatus(id=trainer_uuid,
+                                name=trainer_name,
+                                state=self.state,
+                                errors={},
+                                uptime=self.training_uptime,
+                                progress=self.general_progress)
+
+        status.pretrained_models = self.provided_pretrained_models
+        status.architecture = self.model_architecture
+
+        if self._training:
+            status.errors = self.errors.errors
+            status.context = self.training_context
+
+        return status
+
     async def try_continue_run_if_incomplete(self) -> bool:
         """Tries to continue a training if the last training was not finished.
         """
         if not self.training_active and self.last_training_io.exists():
             self._init_from_last_training()
-            logging.info('found incomplete training, continuing now.')
-            asyncio.get_event_loop().create_task(self._run())
+            logger.info('found incomplete training, continuing now.')
+            self._begin_training_task()
             return True
         return False
 
@@ -185,29 +195,34 @@ class TrainerLogicGeneric(ABC):
         """
         self._training = self.last_training_io.load()
         assert self._training is not None and self._training.training_folder is not None, 'could not restore training folder'
+        logger.info('restored training: \n%s', self._training)
         self._active_training_io = ActiveTrainingIO(
             self._training.training_folder, self.node.loop_communicator, self._training.context)
 
-    async def begin_training(self, organization: str, project: str, details: Dict) -> None:
+    async def begin_training(self, organization: str, project: str, training_config: Dict) -> None:
         """Called on `begin_training` event from the Learning Loop.
         """
-        self._init_new_training(Context(organization=organization, project=project), details)
-        asyncio.get_event_loop().create_task(self._run())
+        self._init_new_training(Context(organization=organization, project=project), training_config)
+        self._begin_training_task()
 
-    def _init_new_training(self, context: Context, details: Dict) -> None:
+    def _begin_training_task(self) -> None:
+        # NOTE: Task object is used to potentially cancel the task
+        self.training_task = asyncio.get_event_loop().create_task(self._run())
+
+    def _init_new_training(self, context: Context, training_config: Dict) -> None:
         """Called on `begin_training` event from the Learning Loop.
-        Note that details needs the entries 'categories' and 'training_number',
+        Note that training_config needs the entries 'categories', 'model_variant' and 'training_number',
         but also the hyperparameter entries.
+        'base_model_uuid' is optional if the training is continued from a previous training.
         """
         project_folder = create_project_folder(context)
         if not self._environment_vars.keep_old_trainings:
             delete_all_training_folders(project_folder)
-        self._training = generate_training(project_folder, context)
-        self._training.set_values_from_data(details)
+        self._training = Training.generate_training(project_folder, context, training_config)
 
         self._active_training_io = ActiveTrainingIO(
             self._training.training_folder, self.node.loop_communicator, context)
-        logging.info(f'new training initialized: {self._training}')
+        logger.info('new training initialized: %s', self._training)
 
     async def _run(self) -> None:
         """Called on `begin_training` event from the Learning Loop. 
@@ -215,22 +230,24 @@ class TrainerLogicGeneric(ABC):
         """
         self.errors.reset_all()
         try:
-            self.training_task = asyncio.get_running_loop().create_task(self._training_loop())
-            await self.training_task  # NOTE: Task object is used to potentially cancel the task
+            await self._training_loop()
         except asyncio.CancelledError:
             if not self.shutdown_event.is_set():
-                logging.info('training task was cancelled but not by shutdown event')
+                logger.info('CancelledError in _run - training task was cancelled but not by shutdown event')
                 self.training.training_state = TrainerState.ReadyForCleanup
                 self.last_training_io.save(self.training)
                 await self._clear_training()
-        except Exception as e:
-            logging.exception(f'Error in train: {e}')
+                self._may_restart()
+            else:
+                logger.info('CancelledError in _run - shutting down')
+        except Exception:
+            logger.exception('(Ignored) exception in trainer_logic._run:')
 
     # ---------------------------------------- TRAINING STATES ----------------------------------------
 
     async def _training_loop(self) -> None:
         """Cycle through the training states until the training is finished or 
-        an asyncio.CancelledError is raised.
+        a critical error occurs (asyncio.CancelledError or CriticalError).
         """
         assert self.training_active
 
@@ -245,20 +262,27 @@ class TrainerLogicGeneric(ABC):
             elif tstate == TrainerState.TrainModelDownloaded:  # -> TrainingRunning -> TrainingFinished
                 await self._perform_state('run_training', TrainerState.TrainingRunning, TrainerState.TrainingFinished, self._train)
             elif tstate == TrainerState.TrainingFinished:  # -> ConfusionMatrixSyncing -> ConfusionMatrixSynced
-                await self._perform_state('sync_confusion_matrix', TrainerState.ConfusionMatrixSyncing, TrainerState.ConfusionMatrixSynced, self._sync_confusion_matrix)
+                await self._perform_state('sync_confusion_matrix', TrainerState.ConfusionMatrixSyncing, TrainerState.ConfusionMatrixSynced, self._sync_training)
             elif tstate == TrainerState.ConfusionMatrixSynced:  # -> TrainModelUploading -> TrainModelUploaded
                 await self._perform_state('upload_model', TrainerState.TrainModelUploading, TrainerState.TrainModelUploaded, self._upload_model)
             elif tstate == TrainerState.TrainModelUploaded:  # -> Detecting -> Detected
                 await self._perform_state('detecting', TrainerState.Detecting, TrainerState.Detected, self._do_detections)
             elif tstate == TrainerState.Detected:  # -> DetectionUploading -> ReadyForCleanup
                 await self._perform_state('upload_detections', TrainerState.DetectionUploading, TrainerState.ReadyForCleanup, self.active_training_io.upload_detetions)
-            elif tstate == TrainerState.ReadyForCleanup:  # -> RESTART or TrainingFinished
+            elif tstate == TrainerState.ReadyForCleanup:  # -> Idle (RESTART or _training = None)
                 await self._clear_training()
                 self._may_restart()
 
     async def _perform_state(self, error_key: str, state_during: TrainerState, state_after: TrainerState, action: Callable[[], Coroutine], reset_early=False):
+        '''
+        Perform a training state and handle errors.
+        - If the loop sends a StopTraining event, this will raise a CancelledError.
+        - States can raise a CriticalError indicating that there is no point in retrying the state.
+        - If any other error occurs, the error is stored in the errors object and the state is reset to the previous state.
+        '''
+
         await asyncio.sleep(0.1)
-        logging.info(f'Performing state: {state_during}')
+        logger.info('Performing state: %s', state_during)
         previous_state = self.training.training_state
         self.training.training_state = state_during
         await asyncio.sleep(0.1)
@@ -266,21 +290,34 @@ class TrainerLogicGeneric(ABC):
             self.errors.reset(error_key)
 
         try:
-            if await action():
-                logging.error('Something went really bad.. cleaning up')
-                state_after = TrainerState.ReadyForCleanup
+            await action()
+
         except asyncio.CancelledError:
-            logging.warning(f'CancelledError in {state_during}')
-            raise
-        except Exception as e:
+            if self.shutdown_event.is_set():
+                logger.info('CancelledError in %s - shutdown event set', state_during)
+                raise
+            logger.info('CancelledError in %s - cleaning up', state_during)
+            self.training.training_state = TrainerState.ReadyForCleanup
+        except CriticalError as e:
+            logger.error('CriticalError in %s - Exception: %s', state_during, e)
             self.errors.set(error_key, str(e))
-            logging.exception(f'Error in {state_during} - Exception:')
+            self.training.training_state = TrainerState.ReadyForCleanup
+        except NodeNeedsRestartError:
+            logger.error('Node Restart Requested')
+            sys.exit(0)
+        except Exception as e:
+            print('Error in %s - Exception: %s', state_during, e, flush=True)
+            self.errors.set(error_key, str(e))
+            logger.exception('Error in %s - Exception: %s', state_during, e)
             self.training.training_state = previous_state
+            return
         else:
+            logger.info('Successfully finished state: %s', state_during)
             if not reset_early:
                 self.errors.reset(error_key)
             self.training.training_state = state_after
-            self.last_training_io.save(self.training)
+
+        self.last_training_io.save(self.training)
 
     async def _prepare(self) -> None:
         """Downloads images to the images_folder and saves annotations to training.data.image_data.
@@ -288,51 +325,59 @@ class TrainerLogicGeneric(ABC):
         self.node.data_exchanger.set_context(self.training.context)
         downloader = TrainingsDownloader(self.node.data_exchanger)
         image_data, skipped_image_count = await downloader.download_training_data(self.training.images_folder)
-        assert self.training.data is not None, 'training.data must be set'
-        self.training.data.image_data = image_data
-        self.training.data.skipped_image_count = skipped_image_count
+
+        self.training.image_data = image_data
+        self.training.skipped_image_count = skipped_image_count
 
     async def _download_model(self) -> None:
         """If training is continued, the model is downloaded from the Learning Loop to the training_folder.
         The downloaded model.json file is renamed to base_model.json because a new model.json will be created during training.
         """
-        base_model_uuid = self.training.base_model_uuid_or_name
+        base_model_uuid = self.training.base_model_uuid
+        base_model_uuid_is_none = base_model_uuid is None
+        base_model_uuid_is_valid = is_valid_uuid4(base_model_uuid)
 
-        # TODO this checks if we continue a training -> make more explicit
-        if not base_model_uuid or not is_valid_uuid4(base_model_uuid):
-            logging.info(f'skipping model download. No base model provided (in form of uuid): {base_model_uuid}')
+        if not base_model_uuid_is_none and not base_model_uuid_is_valid:
+            logger.warning(
+                'base model uuid was provided but was not valid (base_model_uuid: %s).\nSkipping download and starting training from scratch.', base_model_uuid)
             return
 
-        logging.info('loading model from Learning Loop')
-        logging.info(f'downloading model {base_model_uuid} as {self.model_format}')
+        if base_model_uuid_is_none:
+            logger.info('No base model provided (base_model_uuid: %s).\nStarting training from scratch.', base_model_uuid)
+            return
+
+        logger.info('loading model from Learning Loop')
+        logger.info('downloading model %s as %s', base_model_uuid, self.model_format)
         await self.node.data_exchanger.download_model(self.training.training_folder, self.training.context, base_model_uuid, self.model_format)
         shutil.move(f'{self.training.training_folder}/model.json',
                     f'{self.training.training_folder}/base_model.json')
 
-    async def _sync_confusion_matrix(self) -> None:
-        """Syncronizes the confusion matrix with the Learning Loop via the update_training endpoint.
+    async def _sync_training(self) -> None:
+        """Syncronizes the training with the Learning Loop via the update_training endpoint.
         NOTE: This stage sets the errors explicitly because it may be used inside the training stage.
         """
         error_key = 'sync_confusion_matrix'
         try:
             new_best_model = self._get_new_best_training_state()
-            if new_best_model and self.training.data:
+            if new_best_model:
                 new_training = TrainingOut(trainer_id=self.node.uuid,
+                                           trainer_name=self.node.name,
                                            confusion_matrix=new_best_model.confusion_matrix,
-                                           train_image_count=self.training.data.train_image_count(),
-                                           test_image_count=self.training.data.test_image_count(),
-                                           hyperparameters=self.hyperparameters_for_state_sync)
+                                           train_image_count=self.training.train_image_count(),
+                                           test_image_count=self.training.test_image_count(),
+                                           hyperparameters=self.hyperparameters_for_state_sync,
+                                           best_epoch=new_best_model.epoch)
                 await asyncio.sleep(0.1)  # NOTE needed for tests.
 
                 result = await self.node.sio_client.call('update_training', (
                     self.training.context.organization, self.training.context.project, jsonable_encoder(new_training)))
                 if isinstance(result,  dict) and result['success']:
-                    logging.info(f'successfully updated training {asdict(new_training)}')
+                    logger.info('successfully updated training %s', asdict(new_training))
                     self._on_metrics_published(new_best_model)
                 else:
                     raise Exception(f'Error for update_training: Response from loop was : {result}')
         except Exception as e:
-            logging.exception('Error during confusion matrix syncronization')
+            logger.exception('Error during confusion matrix syncronization')
             self.errors.set(error_key, str(e))
             raise
         self.errors.reset(error_key)
@@ -341,21 +386,23 @@ class TrainerLogicGeneric(ABC):
         """Uploads the latest model to the Learning Loop.
         """
         new_model_uuid = await self._upload_model_return_new_model_uuid(self.training.context)
-        if new_model_uuid is None:
-            self.training.training_state = TrainerState.ReadyForCleanup
-            logging.error('could not upload model - maybe training failed.. cleaning up')
-        logging.info(f'Successfully uploaded model and received new model id: {new_model_uuid}')
+        logger.info('Successfully uploaded model and received new model id: %s', new_model_uuid)
         self.training.model_uuid_for_detecting = new_model_uuid
 
-    async def _upload_model_return_new_model_uuid(self, context: Context) -> Optional[str]:
+    async def _upload_model_return_new_model_uuid(self, context: Context) -> str:
         """Upload model files, usually pytorch model (.pt) hyp.yaml and the converted .wts file.
         Note that with the latest trainers the conversion to (.wts) is done by the trainer.
         The conversion from .wts to .engine is done by the detector (needs to be done on target hardware).
-        Note that trainer may train with different classes, which is why we send an initial model.json file."""
+        Note that trainer may train with different classes, which is why we send an initial model.json file.
+
+        :return: The new model UUID.
+        :raise CriticalError: If the latest model files cannot be obtained.
+        :raise Exception: If the response for the model upload attempt is invalid.
+        """
 
         files = await self._get_latest_model_files()
         if files is None:
-            return None
+            raise CriticalError('Could not get latest model files. Training might have failed.')
 
         if isinstance(files, List):
             files = {self.model_format: files}
@@ -369,18 +416,19 @@ class TrainerLogicGeneric(ABC):
             assert len([f for f in _files if 'model.json' in f]) == 1, "model.json must be included exactly once"
 
             model_uuid = await self.node.data_exchanger.upload_model_get_uuid(context, _files, self.training.training_number, file_format)
-            if model_uuid is None:
-                return None
 
             already_uploaded_formats.append(file_format)
             self.active_training_io.save_model_upload_progress(already_uploaded_formats)
 
+        if not model_uuid:
+            raise Exception('Invalid response for model upload attempt')
+
         return model_uuid
 
     def _dump_categories_to_json(self) -> str:
         """Dumps the categories to a json file and returns the path to the file.
         """
-        content = {'categories': [asdict(c) for c in self.training_data.categories], } if self.training_data else None
+        content = {'categories': [asdict(c) for c in self._training.categories], } if self._training else None
         json_path = '/tmp/model.json'
         with open(json_path, 'w') as f:
             json.dump(content, f)
@@ -411,23 +459,23 @@ class TrainerLogicGeneric(ABC):
         if not self.training_active:
             return
         if self.training_task:
-            logging.info('cancelling training task')
+            logger.info('cancelling training task')
             if self.training_task.cancel():
                 try:
                     await self.training_task
                 except asyncio.CancelledError:
                     pass
-                logging.info('cancelled training task')
+                logger.info('cancelled training task')
                 self._may_restart()
 
     def _may_restart(self) -> None:
         """If the environment variable RESTART_AFTER_TRAINING is set, the trainer will restart after a training.
         """
         if self._environment_vars.restart_after_training:
-            logging.info('restarting')
+            logger.info('restarting')
             sys.exit(0)
         else:
-            logging.info('not restarting')
+            logger.info('not restarting')
     # ---------------------------------------- ABSTRACT METHODS ----------------------------------------
 
     @abstractmethod
@@ -450,12 +498,13 @@ class TrainerLogicGeneric(ABC):
 
     @abstractmethod
     def _get_new_best_training_state(self) -> Optional[TrainingStateData]:
-        """Is called frequently by `_sync_confusion_matrix` to check if a new "best" model is availabe.
+        """Is called frequently by `_sync_training` during training to check if a new "best" model is availabe.
         Returns None if no new model could be found. Otherwise TrainingStateData(confusion_matrix, meta_information).
         `confusion_matrix` contains a dict of all classes:
             - The classes must be identified by their uuid, not their name.
             - For each class a dict with tp, fp, fn is provided (true positives, false positives, false negatives).
         `meta_information` can hold any data which is helpful for self._on_metrics_published to store weight file etc for later upload via self.get_model_files
+        `epoch` is the epoch number of the best model.
         """
         raise NotImplementedError
 
